@@ -1,42 +1,36 @@
-# vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
-
-# Copyright 2016-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# SPDX-FileCopyrightText: Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
-# This file is part of qutebrowser.
-#
-# qutebrowser is free software: you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
-#
-# qutebrowser is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# SPDX-License-Identifier: GPL-3.0-or-later
 
 """Showing messages above the statusbar."""
 
-import typing
+from typing import MutableSequence, Optional
 
-from PyQt5.QtCore import pyqtSlot, pyqtSignal, QTimer, Qt
-from PyQt5.QtWidgets import QWidget, QVBoxLayout, QLabel, QSizePolicy
+from qutebrowser.qt.core import pyqtSlot, pyqtSignal, QTimer, Qt
+from qutebrowser.qt.widgets import QWidget, QVBoxLayout, QLabel, QSizePolicy
 
 from qutebrowser.config import config, stylesheet
-from qutebrowser.utils import usertypes
+from qutebrowser.utils import usertypes, message
 
 
 class Message(QLabel):
 
     """A single error/warning/info message."""
 
-    def __init__(self, level, text, replace, parent=None):
+    def __init__(
+            self,
+            level: usertypes.MessageLevel,
+            text: str,
+            replace: Optional[str],
+            text_format: Qt.TextFormat,
+            parent: QWidget = None,
+    ) -> None:
         super().__init__(text, parent)
         self.replace = replace
-        self.setAttribute(Qt.WA_StyledBackground, True)
+        self.level = level
+        self.setAttribute(Qt.WidgetAttribute.WA_StyledBackground, True)
         self.setWordWrap(True)
+        self.setTextFormat(text_format)
         qss = """
             padding-top: 2px;
             padding-bottom: 2px;
@@ -67,6 +61,31 @@ class Message(QLabel):
             raise ValueError("Invalid level {!r}".format(level))
         stylesheet.set_register(self, qss, update=False)
 
+    @staticmethod
+    def _text_format(info: message.MessageInfo) -> Qt.TextFormat:
+        """The Qt.TextFormat to use based on the given MessageInfo."""
+        return Qt.TextFormat.RichText if info.rich else Qt.TextFormat.PlainText
+
+    @classmethod
+    def from_info(cls, info: message.MessageInfo, parent: QWidget = None) -> "Message":
+        return cls(
+            level=info.level,
+            text=info.text,
+            replace=info.replace,
+            text_format=cls._text_format(info),
+            parent=parent,
+        )
+
+    def update_from_info(self, info: message.MessageInfo) -> None:
+        """Update the text from the given info.
+
+        Both the message this gets called on and the given MessageInfo need to have
+        the same level.
+        """
+        assert self.level == info.level, (self, info)
+        self.setTextFormat(self._text_format(info))
+        self.setText(info.text)
+
 
 class MessageView(QWidget):
 
@@ -76,17 +95,17 @@ class MessageView(QWidget):
 
     def __init__(self, parent=None):
         super().__init__(parent)
-        self._messages = []  # type: typing.MutableSequence[Message]
+        self._messages: MutableSequence[Message] = []
         self._vbox = QVBoxLayout(self)
         self._vbox.setContentsMargins(0, 0, 0, 0)
         self._vbox.setSpacing(0)
-        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
+        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
 
         self._clear_timer = QTimer()
         self._clear_timer.timeout.connect(self.clear_messages)
         config.instance.changed.connect(self._set_clear_timer_interval)
 
-        self._last_text = None
+        self._last_info = None
 
     @config.change_filter('messages.timeout')
     def _set_clear_timer_interval(self):
@@ -108,24 +127,29 @@ class MessageView(QWidget):
         for widget in self._messages:
             self._remove_message(widget)
         self._messages = []
-        self._last_text = None
+        self._last_info = None
         self.hide()
         self._clear_timer.stop()
 
-    @pyqtSlot(usertypes.MessageLevel, str, bool)
-    def show_message(self, level, text, replace=False):
+    @pyqtSlot(message.MessageInfo)
+    def show_message(self, info: message.MessageInfo) -> None:
         """Show the given message with the given MessageLevel."""
-        if text == self._last_text:
+        if info == self._last_info:
             return
 
-        if replace and self._messages and self._messages[-1].replace:
-            self._remove_message(self._messages.pop())
+        if info.replace is not None:
+            existing = [msg for msg in self._messages if msg.replace == info.replace]
+            if existing:
+                assert len(existing) == 1, existing
+                existing[0].update_from_info(info)
+                self.update_geometry.emit()
+                return
 
-        widget = Message(level, text, replace=replace, parent=self)
+        widget = Message.from_info(info)
         self._vbox.addWidget(widget)
         widget.show()
         self._messages.append(widget)
-        self._last_text = text
+        self._last_info = info
         self.show()
         self.update_geometry.emit()
         if config.val.messages.timeout != 0:
@@ -134,5 +158,5 @@ class MessageView(QWidget):
 
     def mousePressEvent(self, e):
         """Clear messages when they are clicked on."""
-        if e.button() in [Qt.LeftButton, Qt.MiddleButton, Qt.RightButton]:
+        if e.button() in [Qt.MouseButton.LeftButton, Qt.MouseButton.MiddleButton, Qt.MouseButton.RightButton]:
             self.clear_messages()
